from pprint import pprint

from typing import List, Union, Optional  # Callable, Iterator, Optional
import warnings  # to catch np.polyfit -->> RankWarning: Polyfit may be poorly conditioned

import numpy as np
import scipy.signal

import sanpy

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


def getMedian_Filter(y: np.ndarray, windowPnts: int = 5, verbose=False):
    """Get median filtered version of y using scipy.signal.medfilt2d"""
    if verbose:
        logger.info("")
    filtered = scipy.signal.medfilt2d(y, windowPnts)
    return filtered


def getSavitzkyGolay_Filter(y: np.ndarray, pnts: int = 5, poly: int = 2, verbose=False):
    """Get SavitzkyGolay filtered version of y using scipy.signal.savgol_filter"""
    if verbose:
        logger.info("")
    filtered = scipy.signal.savgol_filter(y, pnts, poly, mode="nearest", axis=0)
    return filtered


def getFirstDerivative(y):
    """Get the first derivative from a time-series.

    Args:
            y (numpy.ndarr)
    """
    firstDerivative = np.diff(y, axis=0)
    firstDerivative = np.insert(firstDerivative, 0, np.nan)
    return firstDerivative


def getSecondDerivative(y):
    """Get the second derivative from a time-series.

    Args:
            y (numpy.ndarray)
    """
    firstDerivative = getFirstDerivative(y)
    secondDerivative = np.diff(firstDerivative, axis=0)
    secondDerivative = np.insert(secondDerivative, 0, np.nan)
    return secondDerivative


def getThreshold_firstDerivative(y, threshold: float, firstDerivative=None):
    """Find threshold crossings in the first derivative."""
    if firstDerivative is None:
        firstDerivative = getFirstDerivative(y)

    Is = np.where(firstDerivative > threshold)[0]
    Is = np.concatenate(([0], Is))
    Ds = Is[:-1] - Is[1:] + 1
    thresholdPoints = Is[np.where(Ds)[0] + 1]
    return thresholdPoints


def getThreshold_vm(y, threshold: float):
    """Find threshold crossings in Vm (membrane potential)."""
    Is = np.where(y > threshold)[0]
    Is = np.concatenate(([0], Is))
    Ds = Is[:-1] - Is[1:] + 1
    thresholdPoints = Is[np.where(Ds)[0] + 1]
    return thresholdPoints


def reduceByRefractory(spikeTimes: List[int], refractoryPnts: int):
    """If there are fast-spikes, throw-out the second one.

    Args:
    spikeTimes: list of spike time threshold crossing
    refractoryPnts:
    """

    # refractory_ms = 20 #10 # remove spike [i] if it occurs within refractory_ms of spike [i-1]

    spikeTimes = spikeTimes.copy()  # make a copy of the list, we are modifying it

    lastGood = 0  # first spike [0] will always be good, there is no spike [i-1]
    for i in range(len(spikeTimes)):
        if i == 0:
            # first spike is always good
            continue
        dPoints = spikeTimes[i] - spikeTimes[lastGood]
        if dPoints < refractoryPnts:
            # remove spike time [i]
            spikeTimes[i] = 0
        else:
            # spike time [i] was good
            lastGood = i

    # regenerate spikeTimes0 by throwing out any spike time that does not pass 'if spikeTime'
    # spikeTimes[i] that were set to 0 above (they were too close to the previous spike)
    newSpikeTimes = [spikeTime for spikeTime in spikeTimes if spikeTime]
    return newSpikeTimes


def refineWithDerivative(
    y: np.ndarray,
    spikeTimes: List[int],
    dvdtPreWindow_pnts: int,
    firstDeriv=None,
    secondDeriv=None,
    verbose=True,
):
    """Refine spike times by looking for zero crossing in first derivative.

    Additionally, look for peak in second derivative.

    Args:
            y (np.ndarrya) Signal to examine, usually membrane potential
            spikeTimes (list[int]):
            firstDeriv (np.ndarry): Specify this to speed execution time
            secondDeriv (np.ndarry): Specify this to speed execution time

    TODO: pass (prePnts, postPnts) as parameter.

    See: refineWithPercentOfMax
    """

    if firstDeriv is None:
        firstDeriv = getFirstDerivative(y)
    if secondDeriv is None:
        secondDeriv = getSecondDerivative(y)

    n = len(spikeTimes)
    footPntList = [None] * n

    for idx, footPnt in enumerate(spikeTimes):
        lastCrossingPnt = footPnt
        # move forwared a bit in case we are already in a local minima ???
        # footPnt += 2  # TODO: add as param

        preStart = footPnt - dvdtPreWindow_pnts
        preClip = firstDeriv[preStart:footPnt]

        zero_crossings = np.where(np.diff(np.sign(preClip)))[
            0
        ]  # find where derivative flips sign (crosses 0)
        if len(zero_crossings) == 0:
            foundFoot = False
            if verbose:
                logger.error(
                    f"  no foot for peak {idx} at pnt {footPnt} ... did not find zero crossings"
                )
        else:
            foundFoot = True
            lastCrossingPnt = preStart + zero_crossings[-1]

        footPntList[idx] = lastCrossingPnt  # was this and worked, a bit too early

        # find peak in second derivative
        if foundFoot:
            preStart2 = lastCrossingPnt
            postPnts = dvdtPreWindow_pnts
            footPnt2 = preStart2 + postPnts
            preClip2 = secondDeriv[preStart2:footPnt2]
            peakPnt2 = np.argmax(preClip2)
            peakPnt2 += preStart2
            #
            footPntList[idx] = peakPnt2

    return footPntList


def refineWithPercentOfMax(
    y, spikeTimes, dvdt_percentOfMax, firstDeriv=None, verbose=False
):
    """Backup spike times by looking for a percent of maximum in the first derivative.

    See: refineWithDerivative
    """
    window_pnts = 30

    if firstDeriv is None:
        firstDeriv = getFirstDerivative(y)

    newSpikeTimes = []
    spikeErrorList = []

    for idx, spikeTime in enumerate(spikeTimes):
        preDerivClip = firstDeriv[spikeTime - window_pnts : spikeTime]  # backwards
        postDerivClip = firstDeriv[spikeTime : spikeTime + window_pnts]  # forwards

        if len(preDerivClip) == 0:
            logger.error("")
            print(
                "FIX ERROR: spikeDetect_dvdt()",
                "spike",
                idx,
                "at pnt",
                spikeTime,
                "window_pnts:",
                window_pnts,
                "len(preDerivClip)",
                len(preDerivClip),
            )  # preDerivClip = np.flip(preDerivClip)

        # look for % of max in dvdt
        try:
            peakPnt = np.argmax(postDerivClip)
            peakPnt += spikeTime
            peakVal = firstDeriv[peakPnt]

            percentMaxVal = (
                peakVal * dvdt_percentOfMax
            )  # value we are looking for in dv/dt
            preDerivClip = np.flip(preDerivClip)  # backwards
            tmpWhere = np.where(preDerivClip < percentMaxVal)
            tmpWhere = tmpWhere[0]  # we flipped it
            if len(tmpWhere) > 0:
                threshPnt2 = np.where(preDerivClip < percentMaxVal)[0][0]
                threshPnt2 = (spikeTime) - threshPnt2
                # threshPnt2 -= 1 # backup by 1 pnt
                newSpikeTimes.append(threshPnt2)
                spikeErrorList.append(None)

            else:
                errorType = "dvdt Percent"
                errStr = f"Did not find dvdt_percentOfMax: percent={dvdt_percentOfMax} peak dV/dt is {round(peakVal,2)}"
                eDict = getErrorDict(
                    idx, spikeTime, errorType, errStr
                )  # spikeTime is in pnts
                if verbose:
                    logger.error(eDict)
                spikeErrorList.append(eDict)
                # always append, do not REJECT spike if we can't find % in dv/dt
                newSpikeTimes.append(spikeTime)
        except (IndexError, ValueError) as e:
            print(
                "   FIX ERROR: refineWithPercentOfMax() looking for dvdt_percentOfMax"
            )
            print("	  ", "IndexError for spike", idx, spikeTime)
            print("	  ", e)
            # always append, do not REJECT spike if we can't find % in dv/dt
            newSpikeTimes.append(spikeTime)
            spikeErrorList.append(None)

    return newSpikeTimes, spikeErrorList


def keepPeaksAbove(y, spikeTimes, onlyPeaksAbove_mV: float, verbose=False):
    """Keep peaks that are aboe a threshold.

    Args:
            onlyPeaksAbove_mV (float): Keep peaks above this threshold (discard peaks below).
    """
    if onlyPeaksAbove_mV is None:
        return spikeTimes
    goodSpikeTimes = [
        spikeTime for spikeTime in spikeTimes if y[spikeTime] > onlyPeaksAbove_mV
    ]
    if verbose:
        logger.info("reduce peaks from {len(spikeTimes)} to len(goodSpikeTimes})")
    return goodSpikeTimes


def keepPeaksBelow(y, spikeTimes, onlyPeaksBelow_mV: float, verbose=False):
    """Keep peaks that are below a threshold.

    Args:
            onlyPeaksBelow_mV (float): Keep peaks below this threshold (discard peaks above).
    """
    if onlyPeaksBelow_mV is None:
        return spikeTimes
    goodSpikeTimes = [
        spikeTime for spikeTime in spikeTimes if y[spikeTime] < onlyPeaksBelow_mV
    ]
    if verbose:
        logger.info("reduce peaks from {len(spikeTimes)} to len(goodSpikeTimes})")
    return goodSpikeTimes


def getPeaks(y, spikeTimes, peakWindow_pnts: int):
    """Get the peak following the threshold crossing.

    Args:
            peakWindow_pnts (int): Window to look for peaks, after each spikeTime.
    """
    peakPnts = []
    for spikeTime in spikeTimes:
        peakPnt = np.argmax(y[spikeTime : spikeTime + peakWindow_pnts])
        peakPnt += spikeTime
        peakPnts.append(peakPnt)
    return peakPnts


def getHalfWidth(
    y,
    spikePnts,
    spikePeakPnts,
    hwWindowPnts: int,
    dataPointsPerMs,
    halfHeights=[10, 20, 50, 80, 90],
    verbose=False,
):
    """Get half-width for a list of spikes."""
    hwList = [None] * len(spikePnts)
    hwErrorList = [None] * len(spikePnts)
    for idx, spikePnt in enumerate(spikePnts):
        peakPnt = spikePeakPnts[idx]
        hwList[idx], hwErrorList[idx] = getHalfWidth_(
            y,
            spikePnt,
            peakPnt,
            idx,
            hwWindowPnts,
            dataPointsPerMs,
            halfHeights=halfHeights,
            verbose=verbose,
        )

    return hwList, hwErrorList


def getHalfWidth_(
    vm,
    thresholdPnt,
    peakPnt,
    peakIdx: int,
    hwWindowPnts: int,
    dataPointsPerMs: Optional[int] = None,
    halfHeights=[10, 20, 50, 80, 90],
    verbose=False,
):
    """
    Get half-widhts for one spike.

    Args:
            vm ():
            thresholdPnt (int): AP threshold crossing
            peakPnt (int): AP peak
            peakIdx (int): The peak index within all peaks. Needed to trigger a proper error
            hwWindowPnts (int): Window to look after peakPnt for falling vm
            dataPointsPerMs (int):
            halfHeightList (list): List of half-height [10,20,50,80,90]
    """

    # halfWidthWindow_ms = None
    # if dataPointsPerMs is not None:
    # 	halfWidthWindow_ms = hwWindowPnts / dataPointsPerMs

    thresholdVal = vm[thresholdPnt]
    peakVal = vm[peakPnt]
    spikeHeight = peakVal - thresholdVal

    spikeSecond = thresholdPnt / dataPointsPerMs / 1000
    peakSec = peakPnt / dataPointsPerMs / 1000

    widthDictList = []
    errorList = []

    # clear out any existing list
    # spikeDict[iIdx]['widths'] = []

    tmpErrorType = None
    for j, halfHeight in enumerate(halfHeights):
        # halfHeight in [20, 50, 80]

        # search rising/falling phase of vm for this vm
        thisVm = thresholdVal + spikeHeight * (halfHeight * 0.01)

        # todo: logic is broken, this get over-written in following try
        widthDict = {
            "halfHeight": halfHeight,
            "risingPnt": None,
            "fallingPnt": None,
            "widthPnts": None,
            "widthMs": float("nan"),
        }
        # widthMs = float('nan')
        currentErrorType = "no error"
        try:
            postRange = vm[peakPnt : peakPnt + hwWindowPnts]
            fallingPnt = np.where(postRange < thisVm)[0]  # less than
            if len(fallingPnt) == 0:
                # no falling pnts found within hwWindowPnts
                currentErrorType = "falling point"
                raise IndexError
            fallingPnt = fallingPnt[0]  # first falling point
            fallingPnt += peakPnt
            fallingVal = vm[fallingPnt]

            # use the post/falling to find pre/rising
            preRange = vm[thresholdPnt:peakPnt]
            risingPnt = np.where(preRange > fallingVal)[0]  # greater than
            if len(risingPnt) == 0:
                currentErrorType = "rising point"
                raise IndexError
            risingPnt = risingPnt[0]  # first rising point
            risingPnt += thresholdPnt

            # actual width (pnts, ms)
            widthPnts = fallingPnt - risingPnt
            if dataPointsPerMs is not None:
                widthMs = widthPnts / dataPointsPerMs
            else:
                widthMS = None
            # assign
            widthDict["halfHeight"] = halfHeight
            widthDict["risingPnt"] = risingPnt
            widthDict["fallingPnt"] = fallingPnt
            widthDict["widthPnts"] = widthPnts
            widthDict["widthMs"] = widthMs  # can be none

            # widthDictList.append(None)
            errorList.append(None)

        except IndexError as e:
            errorType = "Spike Width"
            errorStr = (
                f'Half width {halfHeight} error in "{currentErrorType}" '
                f"with hwWindowPnts:{hwWindowPnts} "
                f"searching for Vm:{round(thisVm,2)} from peak sec {round(peakSec,2)}"
            )

            eDict = getErrorDict(
                peakIdx, thresholdPnt, errorType, errorStr
            )  # spikeTime is in pnts
            errorList.append(eDict)
            if verbose:
                logger.error(
                    f"peakIdx:{peakIdx} spikeSecond:{spikeSecond} halfHeight:{halfHeight} eDict:{eDict}"
                )

        #
        # spikeDict[iIdx]['widths_'+str(halfHeight)] = widthMs
        widthDictList.append(widthDict)

    #
    return widthDictList, errorList


def getMaximumDepolaizingPotential(
    y, spikeTimes: List[int], mdp_pre_pnts: int, avgWindow_pnts: int, verbose=False
):
    """Get the maximum depolaizing potential between spikes.

    This is a very 'cardiac' measure and is often difficulat to interpret if
    cells do not have a 'ventricular' or rhythmic like AP waveform.

    Args:
            y (np.ndarray):
            spikeTimes (list[int]):
            mdp_pre_nts (int)
            avgWindow_pnts (int)
    """

    preMinPntList = [None] * len(spikeTimes)
    errorList = [[]] * len(spikeTimes)  # list of spikes, list within each spike

    for idx, spikePnt in enumerate(spikeTimes):
        # other algorithms look between spike[i-1] and spike[i]
        # here we are looking in a predefined window
        startPnt = spikeTimes[idx] - mdp_pre_pnts

        if startPnt < 0:
            errorType = "mdp"
            detailStr = "Went past start of recording, setting start to 0"
            errorDict = getErrorDict(idx, spikePnt, errorType, detailStr)
            errorList[idx].append(errorDict)
            startPnt = 0

        preRange = y[startPnt : spikeTimes[idx]]
        preMinPnt = np.argmin(preRange)
        preMinPnt += startPnt
        # the pre min is actually an average around the real minima
        avgRange = y[preMinPnt - avgWindow_pnts : preMinPnt + avgWindow_pnts]
        preMinVal = np.average(avgRange)

        # search backward from spike to find when vm reaches preMinVal (avg)
        # this could be a very long time if there is little to know hyper/re polarization
        # between spikes (like in an excitatory neuron)
        preRange = y[preMinPnt : spikeTimes[idx]]
        preRange = np.flip(preRange)  # we want to search backwards from peak
        try:
            preMinPnt2 = np.where(preRange < preMinVal)[0][0]
            preMinPnt = spikeTimes[idx] - preMinPnt2

            preMinPntList[idx] = preMinPnt
            # spikeDict[iIdx]['preMinVal'] = preMinVal

        except IndexError as e:
            errorType = "Pre spike min (mdp)"
            errorStr = "Did not find preMinVal: " + str(
                round(preMinVal, 3)
            )  # + ' postRange min:' + str(np.min(postRange)) + ' max ' + str(np.max(postRange))
            eDict = getErrorDict(
                idx, spikeTimes[idx], errorType, errorStr
            )  # spikeTime is in pnts
            errorList[idx].append(eDict)
            if verbose:
                logger.error(f"  spike:{idx} error:{eDict}")
            # preMinPntList.append(None)

    return preMinPntList, errorList


def getEarlyDiastolicDurationRate(
    y,
    spikeTimes: List[int],
    preMinPnts: List[int],
    lowEddRate_warning: float,
    verbose=False,
):
    """Cardiac specific.

    Args:
            spikeTimes:
            preMinPnts (List[int]): Points where we found MDP.
                    See getMaximumDepolaizingPotential()
    """
    #
    # The nonlinear late diastolic depolarization phase was
    # estimated as the duration between 1% and 10% dV/dt
    # linear fit on 10% - 50% of the time from preMinPnt to self.spikeTimes[i]
    startLinearFit = 0.1  # percent of time between pre spike min and AP peak
    stopLinearFit = 0.5  #
    defaultVal = float("nan")

    x = np.arange(start=0, stop=len(y), step=1)

    retSpikeDict = [{}] * len(spikeTimes)
    errorList = [[]] * len(spikeTimes)

    for idx, spikeTime in enumerate(spikeTimes):
        preMinPnt = preMinPnts[idx]  # can be nan
        if preMinPnt is None:
            # TODO: If all are None then our return dict will be empty
            continue
        timeInterval_pnts = spikeTime - preMinPnt
        # taking round() so we always get an integer # points
        preLinearFitPnt0 = preMinPnt + round(timeInterval_pnts * startLinearFit)
        preLinearFitPnt1 = preMinPnt + round(timeInterval_pnts * stopLinearFit)
        preLinearFitVal0 = y[preLinearFitPnt0]
        preLinearFitVal1 = y[preLinearFitPnt1]

        # linear fit before spike
        retSpikeDict[idx]["preLinearFitPnt0"] = preLinearFitPnt0
        retSpikeDict[idx]["preLinearFitPnt1"] = preLinearFitPnt1
        # retSpikeDict[idx]['earlyDiastolicDuration_ms'] = self.pnt2Ms_(preLinearFitPnt1 - preLinearFitPnt0)
        retSpikeDict[idx]["preLinearFitVal0"] = preLinearFitVal0
        retSpikeDict[idx]["preLinearFitVal1"] = preLinearFitVal1

        # a linear fit where 'm,b = np.polyfit(x, y, 1)', m*x+b"
        xFit = x[preLinearFitPnt0:preLinearFitPnt1]
        yFit = y[preLinearFitPnt0:preLinearFitPnt1]

        # sometimes xFit/yFit have 0 length -->> TypeError
        # print(f' {iIdx} preLinearFitPnt0:{preLinearFitPnt0}, preLinearFitPnt1:{preLinearFitPnt1}')
        # print(f'    xFit:{len(xFit)} yFit:{len(yFit)}')

        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                mLinear, bLinear = np.polyfit(
                    xFit, yFit, 1
                )  # m is slope, b is intercept
                retSpikeDict[idx]["earlyDiastolicDurationRate"] = mLinear
                # make an error if edd rate is too low
                if mLinear <= lowEddRate_warning:
                    errorType = "Fit EDD"
                    errorStr = f"Early diastolic duration rate fit - Too low {round(mLinear,3)}<={lowEddRate_warning}"
                    eDict = getErrorDict(
                        idx, spikeTime, errorType, errorStr
                    )  # spikeTime is in pnts
                    errorList[idx].append(eDict)
                    if verbose:
                        print(f"  spike:{idx} error:{eDict}")

            except TypeError as e:
                # catching exception:  expected non-empty vector for x
                # xFit/yFit turn up empty when mdp and TOP points are within 1 point
                retSpikeDict[idx]["earlyDiastolicDurationRate"] = defaultVal
                errorType = "Fit EDD"
                errorStr = "Early diastolic duration rate fit - preMinPnt == spikePnt"
                eDict = getErrorDict(idx, spikeTime, errorType, errorStr)
                errorList[idx].append(eDict)
                if verbose:
                    logger.error(f"  spike:{idx} error:{eDict}")
            except np.RankWarning as e:
                # logger.error('== FIX preLinearFitPnt0/preLinearFitPnt1 RankWarning')
                # logger.error(f'  error is: {e}')
                # print('RankWarning')
                # also throws: RankWarning: Polyfit may be poorly conditioned
                retSpikeDict[idx]["earlyDiastolicDurationRate"] = defaultVal
                errorType = "Fit EDD"
                errorStr = "Early diastolic duration rate fit - RankWarning"
                eDict = getErrorDict(idx, spikeTime, errorType, errorStr)
                errorList[idx].append(eDict)
                if verbose:
                    logger.error(f"  spike:{idx} error:{eDict}")
            except:
                logger.error(
                    f" !!!!!!!! UNKNOWN EXCEPTION DURING EDD LINEAR FIT for spike {idx}"
                )
                retSpikeDict[idx]["earlyDiastolicDurationRate"] = defaultVal
                errorType = "Fit EDD"
                errorStr = "Early diastolic duration rate fit - Unknown Exception"
                eDict = getErrorDict(idx, spikeTime, errorType, errorStr)
                errorList[idx].append(eDict)
                if verbose:
                    logger.error(f"  spike:{idx} error:{eDict}")

    return retSpikeDict, errorList


def getErrorDict(spikeNumber, pnt, type, detailStr):
    """
    Get error dictionary for one spike.
    """
    # sec = self.pnt2Sec_(pnt)  # pnt / self.dataPointsPerMs / 1000
    # sec = round(sec,4)

    eDict = {
        "Spike": spikeNumber,
        "pnt": pnt,
        #'Seconds': sec,
        "Type": type,
        "Details": detailStr,
    }
    return eDict


def peakDetect(sweepY, detectionParams: sanpy.bDetection, dataPointsPerMs: int):
    """Detect peaks in sweepY using specified detectionParams.

    Args:
            sweepY (np.ndarray): The signal to do detection on, usually a current-clamp Vm recording.
            detectionParams (sanpy.bDetection): Class that specifies the time-constants
            dataPointsPerMs (int):
    """

    firstDerivative = getFirstDerivative(sweepY)

    verbose = detectionParams["verbose"]

    #
    # pre-filter
    medianFilter = detectionParams["medianFilter"]  # 0 means no filter
    if medianFilter > 0:
        sweepY = getMedian_Filter(sweepY, medianFilter, verbose)

    SavitzkyGolay_pnts = detectionParams["SavitzkyGolay_pnts"]  # 0 means no filter
    SavitzkyGolay_poly = detectionParams["SavitzkyGolay_poly"]
    if SavitzkyGolay_pnts > 0:
        sweepY = getSavitzkyGolay_Filter(
            sweepY, SavitzkyGolay_pnts, SavitzkyGolay_poly, verbose
        )

    #
    # detect peaks using different algorithms, either (dvdt, mv)
    detectionType = detectionParams["detectionType"]  # in (dvdt, vm)

    if detectionType == "dvdt":
        dvdtThreshold = detectionParams["dvdtThreshold"]
        dvdtThreshold /= 10  # until I scale my derivative to sampling interval (10 pnts per ms is 10x slower)
        spikePnts = getThreshold_firstDerivative(
            sweepY, dvdtThreshold, firstDerivative=firstDerivative
        )
    elif detectionType == "vm":
        mvThreshold = detectionParams["mvThreshold"]
        spikePnts = getThreshold_vm(sweepY, mvThreshold)

    print(f"1 num spikes:{len(spikePnts)} are {spikePnts}")

    #
    # remove fast spikes
    refractory_pnts = detectionParams.getMsValueAsPnt("refractory_ms", dataPointsPerMs)
    spikePnts = reduceByRefractory(spikePnts, refractoryPnts=refractory_pnts)

    # refine spike times by looking in 1st and 2nd derivative
    # in general this is not working for (cardiac myocytes, DRG neurons) ???
    # dvdtPreWindow_pnts = detectionParams.getMsValueAsPnt('dvdtPreWindow_ms', dataPointsPerMs)
    # spikePnts = refineWithDerivative(sweepY, spikePnts, dvdtPreWindow_pnts,
    # 									firstDeriv=firstDerivative, secondDeriv=secondDerivative,
    # 									verbose=verbose)

    #
    # shift spike times to a percent of dvdt
    dvdt_percentOfMax = detectionParams["dvdt_percentOfMax"]
    spikePnts, errorList = refineWithPercentOfMax(
        sweepY,
        spikePnts,
        dvdt_percentOfMax,
        firstDeriv=firstDerivative,
        verbose=verbose,
    )

    #
    # remove based on (onlyPeaksAbove_mV, onlyPeaksBelow_mV)
    onlyPeaksAbove_mV = detectionParams["onlyPeaksAbove_mV"]
    spikePnts = keepPeaksAbove(sweepY, spikePnts, onlyPeaksAbove_mV, verbose=verbose)

    onlyPeaksBelow_mV = detectionParams["onlyPeaksBelow_mV"]
    spikePnts = keepPeaksBelow(sweepY, spikePnts, onlyPeaksBelow_mV, verbose=verbose)

    ###
    # now that we have a good threshold crossing, calculate everything else
    # number in peakPnts will not change
    ###
    print(f"2 num spikes:{len(spikePnts)} are {spikePnts}")

    # peak
    peakWindow_pnts = detectionParams.getMsValueAsPnt("peakWindow_ms", dataPointsPerMs)
    spikePeakPnts = getPeaks(sweepY, spikePnts, peakWindow_pnts)

    # half-width
    halfWidthWindow_pnts = detectionParams.getMsValueAsPnt(
        "halfWidthWindow_ms", dataPointsPerMs
    )
    halfHeights = detectionParams["halfHeights"]

    hwList, hwErrorList = getHalfWidth(
        sweepY,
        spikePnts,
        spikePeakPnts,
        halfWidthWindow_pnts,
        dataPointsPerMs,
        halfHeights=halfHeights,
        verbose=verbose,
    )

    # maximum diastolic potential (MDP) between spikes
    mdp_pre_pnts = detectionParams.getMsValueAsPnt("mdp_ms", dataPointsPerMs)
    avgWindow_pnts = detectionParams.getMsValueAsPnt("avgWindow_ms", dataPointsPerMs)

    mdpList, mdpErrorList = getMaximumDepolaizingPotential(
        sweepY,
        spikePnts,
        mdp_pre_pnts=mdp_pre_pnts,
        avgWindow_pnts=avgWindow_pnts,
        verbose=verbose,
    )

    lowEddRate_warning = detectionParams["lowEddRate_warning"]
    eddRateList, eddRateErrorList = getEarlyDiastolicDurationRate(
        sweepY,
        spikePnts,
        mdpList,
        lowEddRate_warning=lowEddRate_warning,
        verbose=verbose,
    )

    resultsDict = {
        "spikeTimes": spikePnts,
        "spikePeaks": spikePeakPnts,
        "widths": hwList,
    }

    return resultsDict


def peakDetect_ba(ba, sweep, detectionParams):
    """Detect one sweep in a bAnalysis."""
    ba.setSweep(sweep)

    sweepY = ba.sweepY
    dataPointsPerMs = ba.dataPointsPerMs

    resultsDict = peakDetect(sweepY, detectionParams, dataPointsPerMs)

    return resultsDict


def testNeuron():
    path = "/Users/cudmore/data/theanne-griffith/07.20.21/2021_07_20_0013.abf"
    ba = sanpy.bAnalysis(path)
    print(ba)

    detectionPreset = sanpy.bDetection.detectionPresets.neuron
    detectionParams = sanpy.bDetection(detectionPreset=detectionPreset)

    for sweep in range(ba.numSweeps):
        resultsDict = peakDetect_ba(ba, sweep, detectionParams)

        sweepX = ba.sweepX
        sweepX_pnts = np.arange(start=0, stop=sweepX.shape[0], step=1)
        sweepY = ba.sweepY

        Plot(sweepX_pnts, sweepY, resultsDict)
        plt.show()


def _plotHalfWidth(y, hwSpikes, ax):
    """

    Args:
            hwSpikes (list): a list of spikes, each has a list of hw dict.
    """

    for oneSpikeHwList in hwSpikes:
        for oneHw in oneSpikeHwList:
            # oneHw is a dict
            # {'halfHeight': 10, 'risingPnt': 1579, 'fallingPnt': 1590, 'widthPnts': 11, 'widthMs': 1.1}
            # print('  oneHw:', oneHw)

            risingPnt = oneHw["risingPnt"]
            fallingPnt = oneHw["fallingPnt"]

            if risingPnt is None or fallingPnt is None:
                # error, we did not find a rising/falling point
                continue

            yRisingPnt = y[risingPnt]
            yFallingPnt = y[fallingPnt]

            ax.plot(risingPnt, yRisingPnt, "o")
            ax.plot(fallingPnt, yFallingPnt, "o")


def Plot(x, y, resultsDict: dict = {}, showFirstDeriv=True, showSecondDeriv=True):
    if showFirstDeriv:
        firstDerivative = getFirstDerivative(y)
    if showSecondDeriv:
        secondDerivative = getSecondDerivative(y)

    if showFirstDeriv or showSecondDeriv:
        numRows = 2
        yAxisIdx = 1
    else:
        numRows = 1
        yAxisIdx = 0

    fig, axs = plt.subplots(numRows, 1, sharex=True)

    if showFirstDeriv:
        axs[0].plot(x, firstDerivative, "k", linewidth=1)
        axs[0].hlines(0, x[0], x[-1], linestyles="dashed")
    if showSecondDeriv:
        rightAxes = axs[0].twinx()
        rightAxes.plot(x, secondDerivative, "r", linewidth=1)
        rightAxes.hlines(0, x[0], x[-1], linestyles="dashed")

    axs[yAxisIdx].plot(x, y, "k", linewidth=1)

    if resultsDict is not None:
        #
        spikeTimes = resultsDict["spikeTimes"]
        spikeVals = y[spikeTimes]
        axs[yAxisIdx].plot(spikeTimes, spikeVals, "or")
        #
        spikePeaks = resultsDict["spikePeaks"]
        spikePeaksVals = y[spikePeaks]
        axs[yAxisIdx].plot(spikePeaks, spikePeaksVals, "ob")

        _plotHalfWidth(y, resultsDict["widths"], axs[yAxisIdx])
        # for idx, spikeTime in enumerate(spikeTimes):
        # 	# ['widhts'] is a list of dict
        # 	print(f"idx:{idx} width: {resultsDict['widths'][idx]}")

    # plt.show()


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import sanpy

    testNeuron()



from typing import List

from PyQt5 import QtCore, QtGui, QtWidgets

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class errorTableView(QtWidgets.QTableView):
    """
    Display a per spike error table (one row per spike eror)
    """

    signalSelectSpike = QtCore.Signal(object)  # spike number, doZoom

    def __init__(self, parent=None):
        super(errorTableView, self).__init__(parent)

        # self.setFont(QtGui.QFont('Arial', 10))

        self._blockTableUpdate = False

        self.setSortingEnabled(True)

        # self.setSizePolicy(
        #     QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        # )

        _hHeader = self.horizontalHeader()
        _hHeader.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        # _hHeader.setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Stretch)

        # see
        # https://stackoverflow.com/questions/38098763/pyside-pyqt-how-to-make-set-qtablewidget-column-width-as-proportion-of-the-a
        # _hHeader = self.horizontalHeader()       
        # _hHeader.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Stretch)
        # _hHeader.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        # _hHeader.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)

        self.setSelectionBehavior(QtWidgets.QTableView.SelectRows)

        # self.setSelectionMode(QtWidgets.QTableView.SingleSelection)

        self.setEditTriggers(
            QtWidgets.QAbstractItemView.NoEditTriggers
            | QtWidgets.QAbstractItemView.DoubleClicked
        )

        # equally stretchs each columns so that they fit the table's width.
        # self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch);
        self.horizontalHeader().setStretchLastSection(True)

        # removed mar 2023
        # rowHeight = 11
        # fnt = self.font()
        # fnt.setPointSize(rowHeight)
        # self.setFont(fnt)
        # self.verticalHeader().setDefaultSectionSize(rowHeight);

        # self.resizeColumnsToContents()

        self.clicked.connect(self.errorTableClicked)
        self.doubleClicked.connect(self._on_doubleClick)

    def mySelectRows(self, rows: List[int]):
        """Select rows in the table

        TODO: Hold off on this, we need to define a sort proxy. See PyMapManager tables.
        """
        return

        logger.info(f"rows:{rows}")

        # to stop event recursion
        self._blockTableUpdate = True

        selectionModel = self.selectionModel()
        if selectionModel:
            selectionModel.clear()

            if rows:
                indexes = [self.model().index(r, 0) for r in rows]  # [QModelIndex]
                visualRows = [
                    self.proxy.mapFromSource(modelIndex) for modelIndex in indexes
                ]

                mode = (
                    QtCore.QItemSelectionModel.Select | QtCore.QItemSelectionModel.Rows
                )
                [self.selectionModel().select(i, mode) for i in visualRows]

                # scroll so first row in rows is visible
                # TODO (cudmore) does not work if list is filtered
                column = 0
                row = list(rows)[0]
                index = self.model().index(row, column)
                self.scrollTo(
                    index, QtWidgets.QAbstractItemView.PositionAtTop
                )  # EnsureVisible
            else:
                # print('  CLEARING SELECTION')
                self.clearSelection()

        #
        self._blockTableUpdate = False

    def _on_doubleClick(self, index):
        """
        Parameters
        ----------
        index : QtCore.QModelIndex
        """
        row = self.model()._data.index[index.row()]  # take sorting into account
        logger.info(index.row())
        self.errorTableClicked(index, doubleClick=True)

    def errorTableClicked(self, index, doubleClick=False):
        """
        Parameters
        ----------
        index : QtCore.QModelIndex
        """
        row = self.model()._data.index[index.row()]  # take sorting into account

        doZoom = doubleClick
        # modifiers = QtWidgets.QApplication.keyboardModifiers()
        # if modifiers == QtCore.Qt.ShiftModifier:
        #     # zoomm on shift+click
        #     doZoom = True

        try:
            # .loc[i,'col'] gets from index label (correct)
            # .iloc[i,j] gets absolute row (wrong)
            spikeNumber = self.model()._data.loc[row, "Spike"]
        except KeyError as e:
            # for results plugin
            try:
                spikeNumber = self.model()._data.loc[row, "spikeNumber"]
            except KeyError as e:
                logger.warning(f'KeyError looking for column "Spike" or "spikeNumber"')
                return
        spikeNumber = int(spikeNumber)

        dDict = {
            "spikeNumber": spikeNumber,
            "doZoom": doZoom,
        }
        logger.info(f"  -->> emit signalSelectSpike with dDict")
        logger.info(dDict)

        # self._blockSlots = True
        self.signalSelectSpike.emit(dDict)

        # self._blockSlots = False



import os
import pathlib
import json
from json.decoder import JSONDecodeError

import sanpy
import sanpy.interface
import sanpy.interface.sanpy_app as sanpy_app

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class preferences:
    """Class to hold GUI preferences.

    Created and used by sanpy app SanPyWindow.
    """

    def __init__(self, sanpyApp: sanpy_app):
        super().__init__()
        self._sanpyApp = sanpyApp

        # bump this when we change, sanpy app will rebuild if loaded preerences are out of date
        # self._version = 1.2  # increment when we change preferences
        self._version = 1.3  # added set spike
        self._version = 1.4  # converted plugins from list[str] to dict to include externalWindow option
        self._version = 1.5  # adding keys to plugins (l, t, w, h)

        self._maxRecent = 7  # a lucky number
        self._configDict = self.load()

    def __setitem__(self, key, item):
        self._configDict[key] = item

    def __getitem__(self, key):
        return self._configDict[key]

    @property
    def configDict(self):
        return self._configDict

    def addFolder(self, path: str):
        """Add a folder to preferences."""
        logger.info(f"{path}")

        # add if not in recentFolders
        if not path in self.configDict["recentFolders"]:
            self.configDict["recentFolders"].append(path)
            # limit list to last _maxNumUndo
            self.configDict["recentFolders"] = self.configDict["recentFolders"][
                -self._maxRecent :
            ]

        # always set as the most recent folder
        self.configDict["mostRecentFolder"] = path

    def addPlugin(self, pluginName: str, externalWindow: bool, ltwhTuple: tuple):
        """Add plugin to preferences."""
        # was this as list
        # if not pluginName in self.configDict['pluginPanels']:
        #     self.configDict['pluginPanels'].append(pluginName)
        if not pluginName in self.configDict["pluginPanels"].keys():
            self.configDict["pluginPanels"][pluginName] = {
                "externalWindow": externalWindow,
                "l": ltwhTuple[0],
                "t": ltwhTuple[1],
                "w": ltwhTuple[2],
                "h": ltwhTuple[3],
            }

    def removePlugin(self, pluginName: str):
        """Remove plugin from preferences."""
        # was this as list
        # if pluginName in self.configDict['pluginPanels']:
        #     self.configDict['pluginPanels'].remove(pluginName)
        if pluginName in self.configDict["pluginPanels"].keys():
            self.configDict["pluginPanels"].pop(pluginName)

    def getMostRecentFolder(self):
        return self.configDict["mostRecentFolder"]

    def getRecentFolder(self):
        return self.configDict["recentFolders"]

    def preferencesSet(self, key1, key2, val):
        """Set a preference. See `getDefaults()` for key values."""
        try:
            self._configDict[key1][key2] = val

            # actually show hide some widgets
            # if key1=='display' and key2=='showScatter':
            #    if val:
            #        self.myScatterPlotWidget.show()
            #    else:
            #        self.myScatterPlotWidget.hide()

            # if key1=='display' and key2=='showErrors':
            #    if val:
            #        self.myErrorTable.show()
            #    else:
            #        self.myErrorTable.hide()

        except KeyError as e:
            logger.error(f'Did not set preference with keys "{key1}" and "{key2}"')

    def preferencesGet(self, key1, key2):
        """Get a preference. See `getDefaults()` for key values."""
        try:
            return self._configDict[key1][key2]
        except KeyError as e:
            logger.error(f'Did not get preference with keys "{key1}" and "{key2}"')

    def getPreferencesFile(self):
        userPreferencesFolder = sanpy._util._getUserPreferencesFolder()
        optionsFile = pathlib.Path(userPreferencesFolder) / "sanpy_preferences.json"
        return optionsFile

    def load(self):
        """
        Always load preferences from:
            <user>/Documents/SanPy/preferences/sanpy_preferences.json
        """

        preferencesFile = self.getPreferencesFile()

        useDefault = True
        if os.path.isfile(preferencesFile):
            logger.info(f"Loading preferences file")
            logger.info(f"  {preferencesFile}")
            try:
                with open(preferencesFile) as f:
                    loadedJson = json.load(f)
                    loadedVersion = loadedJson["version"]
                    logger.info(f"  loaded preferences version {loadedVersion}")
                    if loadedVersion < self._version:
                        # use default
                        logger.warning(
                            "  older version found, reverting to current defaults"
                        )
                        logger.warning(
                            f"  loadedVersion:{loadedVersion} currentVersion:{self._version}"
                        )
                        pass
                    else:
                        return loadedJson
            except JSONDecodeError as e:
                logger.error(e)
            except TypeError as e:
                logger.error(e)
        if useDefault:
            logger.info(f"Using default options")
            return self.getDefaults()

    def getDefaults(self) -> dict:
        """Get default preferences.

        Be sure to increment self._version when making changes.
        """
        configDict = {}  # OrderedDict()

        configDict["version"] = self._version
        configDict["useDarkStyle"] = True
        configDict[
            "autoDetect"
        ] = True  # FALSE DOES NOT WORK!!!! auto detect on file selection and/or sweep selection

        configDict["recentFolders"] = []
        configDict["mostRecentFolder"] = ""

        configDict["windowGeometry"] = {}
        configDict["windowGeometry"]["x"] = 100
        configDict["windowGeometry"]["y"] = 100
        configDict["windowGeometry"]["width"] = 1000
        configDict["windowGeometry"]["height"] = 1000

        configDict["filePanels"] = {}
        configDict["filePanels"]["File Panel"] = True

        # panels withing detectionWidget -> myDetectionToolbar
        configDict["detectionPanels"] = {}
        configDict["detectionPanels"]["Detection Panel"] = True  # main panel
        configDict["detectionPanels"]["Detection"] = True
        configDict["detectionPanels"]["Display"] = True
        configDict["detectionPanels"]["Plot Options"] = False
        configDict["detectionPanels"]["Set Spikes"] = False

        # plugins to show at startup
        # 20230325 convert from List[str] to dict
        configDict["pluginPanels"] = {}

        # values in detectionWidget -> myDetectionToolbar
        configDict["detect"] = {}
        configDict["detect"]["detectDvDt"] = 20
        configDict["detect"]["detectMv"] = -20

        # display options within detectionWidget -> myDetectionToolbar
        configDict["rawDataPanels"] = {}
        configDict["rawDataPanels"]["plotEveryPoint"] = 10  # not used?
        configDict["rawDataPanels"]["Full Recording"] = True  #
        configDict["rawDataPanels"]["Derivative"] = False  #
        configDict["rawDataPanels"]["DAC"] = False  #

        return configDict

    def save(self):
        preferencesFile = self.getPreferencesFile()

        logger.info(f'Saving preferences file as: "{preferencesFile}"')

        # get the current position and size of the main window
        left, top, width, height = self._sanpyApp.getWindowGeometry()

        self._configDict["windowGeometry"]["x"] = left
        self._configDict["windowGeometry"]["y"] = top
        self._configDict["windowGeometry"]["width"] = width
        self._configDict["windowGeometry"]["height"] = height

        # TODO: extend to a list
        # self._configDict['lastPath'] = self._sanpyApp.path

        # TODO: get window position and size of all open plugins
        # self.configDict['pluginPanels'][pluginName] = {'externalWindow': externalWindow,
        #                                                    'l': ltwhTuple[0],
        #                                                    't': ltwhTuple[1],
        #                                                    'w': ltwhTuple[2],
        #                                                    'h': ltwhTuple[3],
        #                                                    }
        _openSet = self._sanpyApp.myPlugins._openSet
        for _onePlugin in _openSet:
            pluginName = _onePlugin.myHumanName
            ltwhTuple = _onePlugin.getWindowGeometry()
            self.configDict["pluginPanels"][pluginName]["l"] = ltwhTuple[0]
            self.configDict["pluginPanels"][pluginName]["t"] = ltwhTuple[1]
            self.configDict["pluginPanels"][pluginName]["w"] = ltwhTuple[2]
            self.configDict["pluginPanels"][pluginName]["h"] = ltwhTuple[3]

        #
        # save
        with open(preferencesFile, "w") as outfile:
            json.dump(self._configDict, outfile, indent=4, sort_keys=True)

        self._sanpyApp.slot_updateStatus("Saved preferences")



# 20210525
import os, sys
import logging
from logging.handlers import RotatingFileHandler
import pathlib

"""
DEBUG
INFO
WARNING
ERROR
CRITICAL
"""

"""
LOGLEVEL = os.environ.get('LOGLEVEL', 'WARNING').upper()
print('LOGLEVEL:', LOGLEVEL)
print('os.environ')
for k,v in os.environ.items():
    print(f'  {k}: {v}')
sys.exit(1)
"""

# basedir = os.path.dirname(__file__)
# print('sanpyLogger basedir:', basedir)

# from sanpy._util import _getUserPreferencesFolder


def getLoggerFile():
    """Get the path to save the log file.

    If <user>/Documents/SanPy/preferences exists
    else MEIPASS
    """
    # logPath = sanpy._util._getUserPreferencesFolder()
    # logPath = os.path.join(logPath, 'sanpy.log')
    # return logPath

    userPath = pathlib.Path.home()
    userPreferencesFolder = os.path.join(userPath, "Documents", "SanPy", "preferences")
    # print('looking for userPreferencesFolder:', userPreferencesFolder)
    if os.path.isdir(userPreferencesFolder):
        myPath = userPreferencesFolder
        # print('  my path 1:', myPath)
    elif getattr(sys, "frozen", False):
        # running in a bundle (frozen)
        myPath = sys._MEIPASS
        # print('  my path 2:', myPath)
    else:
        # running in a normal Python environment
        # myPath = os.path.dirname(os.path.abspath(__file__))
        myPath = pathlib.Path(__file__).parent.absolute()
        # print('  my path 3:', myPath)

    fileName = "sanpy.log"
    logPath = os.path.join(myPath, fileName)
    return logPath


def get_logger(name, level=logging.DEBUG):
    """ """

    # if getattr(sys, 'frozen', False):
    #     # running in a bundle (frozen)
    #     myPath = sys._MEIPASS
    # else:
    #     # running in a normal Python environment
    #     myPath = pathlib.Path(__file__).parent.absolute()

    # fileName = 'sanpy.log'
    # logPath = os.path.join(myPath, fileName)

    logPath = getLoggerFile()

    # print('logPath:', logPath)

    # Create a custom logger
    logger = logging.getLogger(name)
    logger.setLevel(level)  # abb 20220609

    # abb removed 20220609
    logger.propagate = False  # don't propogate to root (o.w. prints twice)
    # print('   ', logger, 'level:', level)
    if not logger.handlers:
        # print('=== sanpyLogger.get_logger() creating handlers')
        # print('    ', logger.handlers)

        # Create handlers
        c_handler = logging.StreamHandler()
        f_handler = RotatingFileHandler(logPath, maxBytes=500, backupCount=0)
        # f_handler = logging.FileHandler(logPath)

        c_handler.setLevel(level)
        f_handler.setLevel(level)

        # Create formatters and add it to handlers
        consoleFormat = "%(levelname)5s %(name)8s  %(filename)s %(funcName)s() line:%(lineno)d -- %(message)s"
        c_format = logging.Formatter(consoleFormat)

        fileFormat = "%(asctime)s  %(levelname)5s %(name)8s  %(filename)s %(funcName)s() line:%(lineno)d -- %(message)s"
        f_format = logging.Formatter(fileFormat)

        c_handler.setFormatter(c_format)
        f_handler.setFormatter(f_format)

        # Add handlers to the logger
        logger.addHandler(c_handler)
        logger.addHandler(f_handler)
    #
    #
    return logger


# see: https://stackoverflow.com/questions/6234405/logging-uncaught-exceptions-in-python
def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = handle_exception

# This does not seem like a good idea, without it, we get exception calling
# logger.error in handle_exception (above)
logger = get_logger(__name__)


def test():
    logger.error("111")


if __name__ == "__main__":
    logger = get_logger(__name__)
    logger.error("WORKS")

    test()

    # print(1/0)

    """
    a = 5
    b = 0
    try:
        c = a / b
    except Exception as e:
        logger.exception("Exception occurred")
    """



import os
import sys
import time
import pathlib
import shutil

import pandas as pd

# april 5, removed (why did I remove???)
import tables.scripts.ptrepack  # to save compressed .h5 file

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

def listKeys(hdfPath, printData=False):
    """List all keys in h5 file."""
    with pd.HDFStore(hdfPath, mode="r") as store:
        logger.info(f'store: {store}')
        for key in store.keys():
            print("    key:", key, "type:", type(store[key]))
            if printData:
                print(store[key])


def _getTmpHdfFile(hdfPath):
    """Get temporary h5 file to write to.

    We will always then compress with _rebuildHdf.

    Used in analysisDir to delete uuid from h5 file
    """
    if not os.path.isfile(hdfPath):
        logger.error(f"File not found {hdfPath}")
        return

    logger.info("")

    _folder, _file = os.path.split(hdfPath)

    tmpHdfFile = os.path.splitext(_file)[0] + "_tmp.h5"
    tmpHdfPath = pathlib.Path(_folder) / tmpHdfFile

    # the compressed version from the last save
    hdfFile = os.path.splitext(_file)[0] + ".h5"
    hdfFilePath = pathlib.Path(_folder) / hdfFile

    # hdfMode = 'w'
    if os.path.isfile(hdfFilePath):
        logger.info(f"    copying existing hdf file to tmp ")
        logger.info(f"    hdfFilePath {hdfFilePath}")
        logger.info(f"    tmpHdfPath {tmpHdfPath}")
        shutil.copyfile(hdfFilePath, tmpHdfPath)
    else:
        pass
        # compressed file does not exist, just use tmp path
        # print('   does not exist:', hdfFilePath)

    return tmpHdfPath


def _repackHdf(hdfPath):
    _folder, _file = os.path.split(hdfPath)

    # rebuild the file to remove old changes and reduce size
    # tmpHdfFile = os.path.splitext(_file)[0] + '_tmp.h5'
    # tmpHdfPath = pathlib.Path(_folder) / tmpHdfFile

    # always write to a tmp and then repack into hdfPath
    tmpHdfPath = _getTmpHdfFile(hdfPath)

    hdfFile = os.path.splitext(_file)[0] + ".h5"
    hdfPath = pathlib.Path(_folder) / hdfFile
    logger.info(f"Rebuilding h5 to {hdfPath}")

    # can't pass sys.argv a 'PosixPath' from pathlib.Path, needs to be a string
    tmpHdfPath = str(tmpHdfPath)
    hdfPath = str(hdfPath)

    # when calling ptrepack, we need trailing ':' on each src/dst path
    # without this Windows fails to find the file
    _tmpHdfPath = tmpHdfPath + ":"
    _hdfPath = hdfPath + ":"

    # The first item is normally the command line command name (not used)
    sys.argv = ["", "--overwrite", "--chunkshape=auto", _tmpHdfPath, _hdfPath]

    logger.info("    running tables.scripts.ptrepack.main()")
    logger.info(f"    sys.argv: {sys.argv}")
    try:
        tables.scripts.ptrepack.main()

        # delete the temporary (large file)
        logger.info(f"    Deleting tmp file: {tmpHdfPath}")
        os.remove(tmpHdfPath)

        # self.signalApp(f'Saved compressed folder analysis with tables.scripts.ptrepack.main()')

    except FileNotFoundError as e:
        logger.error("tables.scripts.ptrepack.main() failed ... file was not saved")
        logger.error(e)
        # self.signalApp(f'ERROR in tables.scripts.ptrepack.main(): {e}')


def _loadAnalysis(hdfPath):
    """Load all bAnalysis from h5"""
    logger.info(f"hdfPath: {hdfPath}")
    start = time.time()
    numLoaded = 0
    with pd.HDFStore(hdfPath, mode="r") as store:
        for key in store.keys():
            data = store[key]
            if isinstance(data, pd.DataFrame):
                if "_sweepX" in data.columns:
                    ba = sanpy.bAnalysis(fromDf=data)
                    logger.info(f"loaded ba: {ba}")
                    numLoaded += 1
                else:
                    # this is usually the file database
                    print("Dur(s):")
                    print(store[key]["Dur(s)"])
                    # the entire file db df
                    print(data)
    #
    stop = time.time()
    print(f"Loading {numLoaded} bAnalysis took {round(stop-start,3)} seconds.")


if __name__ == "__main__":
    import sanpy.bAnalysis

    hdfPath = "/home/cudmore/Sites/SanPy/data/sanpy_recording_db.h5"
    # listKeys(hdfPath)
    _loadAnalysis(hdfPath)

    # the file befor compression
    """
    hdfPath = '/home/cudmore/Sites/SanPy/data/sanpy_recording_db_tmp.h5'
    loadAnalysis(hdfPath)
    """



import os
import sys
import importlib
import pathlib
import shutil
from typing import List, Union
import uuid

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


def getNewUuid():
    return "t" + str(uuid.uuid4()).replace("-", "_")


def getBundledDir():
    """Get the working directory where user preferences are save.

    This will be source code folder when running from source,
      will be a more freeform folder when running as a frozen app/exe
    """
    if getattr(sys, "frozen", False):
        # we are running in a bundle (frozen)
        bundle_dir = sys._MEIPASS
    else:
        # we are running in a normal Python environment
        bundle_dir = os.path.dirname(os.path.abspath(__file__))
    return bundle_dir


def _module_from_file(module_name: str, file_path: str):
    """

    Args:
        module_name: Is like sanpy.interface.plugins.onePluginFile
        file_path: Full path to onePluginFile source code (onePluginFile.py)
    """
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def addUserPath():
    """Make <user>/Documents/SanPy folder and add it to the Python sys.path

    Returns:
        True: If we made the folder (first time SanPy is running)
    """

    logger.info("")

    madeUserFolder = _makeSanPyFolders()  # make <user>/Documents/SanPy if necc

    userSanPyFolder = _getUserSanPyFolder()

    # if userSanPyFolder in sys.path:
    #     sys.path.remove(userSanPyFolder)

    if not userSanPyFolder in sys.path:
        logger.info(f"Adding to sys.path: {userSanPyFolder}")
        sys.path.append(userSanPyFolder)

    logger.info("sys.path is now:")
    for path in sys.path:
        logger.info(f"    {path}")

    return madeUserFolder


def _getUserDocumentsFolder():
    """Get <user>/Documents folder."""
    userPath = pathlib.Path.home()
    userDocumentsFolder = os.path.join(userPath, "Documents")
    if not os.path.isdir(userDocumentsFolder):
        logger.error(f'Did not find path "{userDocumentsFolder}"')
        logger.error(f'   Using "{userPath}"')
        return userPath
    else:
        return userDocumentsFolder


def _getUserSanPyFolder():
    """Get <user>/Documents/SanPy folder."""
    userDocumentsFolder = _getUserDocumentsFolder()
    sanpyFolder = os.path.join(userDocumentsFolder, "SanPy-User-Files")
    return sanpyFolder


def _old_getUserFolder(folder: str) -> str:
    userSanPyFolder = _getUserSanPyFolder()
    if folder == "plugins":
        theFolder = os.path.join(userSanPyFolder, "plugins")
    elif folder == "analysis":
        theFolder = os.path.join(userSanPyFolder, "analysis")
    elif folder == "preferences":
        theFolder = os.path.join(userSanPyFolder, "preferences")
    else:
        logger.error(f'did not understand folder: "{folder}"')
        return None
    return theFolder


def _getUserFileLoaderFolder():
    userSanPyFolder = _getUserSanPyFolder()
    fileLoaderFolder = os.path.join(userSanPyFolder, "file loaders")
    return fileLoaderFolder


def _getUserPluginFolder():
    userSanPyFolder = _getUserSanPyFolder()
    userPluginFolder = os.path.join(userSanPyFolder, "plugins")
    return userPluginFolder


def _getUserDetectionFolder():
    """Folder of saved user detection presets.

    Each is a json file.
    """
    userSanPyFolder = _getUserSanPyFolder()
    userDetectionFolder = os.path.join(userSanPyFolder, "detection")
    return userDetectionFolder


def _getUserAnalysisFolder():
    """Folder of custom user analysis code."""
    userAnalysisFolder = _getUserSanPyFolder()
    userAnalysisFolder = os.path.join(userAnalysisFolder, "analysis")
    return userAnalysisFolder


def _getUserPreferencesFolder():
    """Folder of SanPy app preferences and logs (user does not modify this."""
    userPreferencesFolder = _getUserSanPyFolder()
    userPreferencesFolder = os.path.join(userPreferencesFolder, "preferences")
    return userPreferencesFolder


def pprint(d: dict):
    for k, v in d.items():
        print(f"  {k}: {v}")


def _makeSanPyFolders():
    """Make <user>/Documents/SanPy-User-Files folder .

    If no Documents folder then make SanPy folder directly in <user> path.
    """
    userDocumentsFolder = _getUserDocumentsFolder()

    madeUserFolder = False

    # main <user>/Documents/SanPy folder
    sanpyFolder = _getUserSanPyFolder()
    if not os.path.isdir(sanpyFolder):
        # first time run
        logger.info(f'Making <user>/SanPy-User-Files folder "{sanpyFolder}"')
        madeUserFolder = True
        #
        # copy entire xxx into <user>/Documents/SanPy
        _bundDir = getBundledDir()
        _srcPath = pathlib.Path(_bundDir) / "_userFiles" / "SanPy-User-Files"
        _dstPath = pathlib.Path(sanpyFolder)
        logger.info(f"    copying folder tree to <user>/Documents/SanPy-User-Folder")
        logger.info(f"    _srcPath:{_srcPath}")
        logger.info(f"    _dstPath:{_dstPath}")
        shutil.copytree(_srcPath, _dstPath)
    else:
        # already exists, make sure we have all sub-folders that are expected
        pass

    return madeUserFolder

def _loadLineScanHeader(path):
    """
    path: full path to tif

    we will load and parse coresponding .txt file

    returns dict:
        numPixels:
        umLength:
        umPerPixel:
        totalSeconds:
    """
    # "X Dimension"	"138, 0.0 - 57.176 [um], 0.414 [um/pixel]"
    # "T Dimension"	"1, 0.000 - 35.496 [s], Interval FreeRun"
    # "Image Size(Unit Converted)"	"57.176 [um] * 35500.000 [ms]"

    # 20220606, adding
    # "Image Size"	"294 * 1000 [pixel]"

    txtFile = os.path.splitext(path)[0] + ".txt"

    if not os.path.isfile(txtFile):
        logger.error(f"did not find file:{txtFile}")
        return None

    theRet = {"tif": path}

    # tif shape is (lines, pixels)
    # theRet['numLines'] = self.tif.shape[1]
    # theRet['numLines'] = tifData.shape[0]

    gotImageSize = False

    with open(txtFile, "r") as fp:
        lines = fp.readlines()
        for line in lines:
            line = line.strip()
            if line.startswith('"X Dimension"'):
                line = line.replace('"', "")
                line = line.replace(",", "")
                # print('loadLineScanHeader:', line)
                # 2 number of pixels in line
                # 5 um length of line
                # 7 um/pixel
                splitLine = line.split()
                for idx, split in enumerate(splitLine):
                    # print('  ', idx, split)
                    if idx == 2:
                        numPixels = int(split)
                        theRet["numPixels"] = numPixels
                    elif idx == 5:
                        umLength = float(split)
                        theRet["umLength"] = umLength
                    elif idx == 7:
                        umPerPixel = float(split)
                        theRet["umPerPixel"] = umPerPixel

            elif line.startswith('"T Dimension"'):
                # "T Dimension"	"1, 0.000 - 35.496 [s], Interval FreeRun"
                line = line.replace('"', "")
                line = line.replace(",", "")
                # print('loadLineScanHeader:', line)
                # 5 total duration of image acquisition (seconds)
                splitLine = line.split()
                for idx, split in enumerate(splitLine):
                    # print('  ', idx, split)
                    if idx == 5:
                        totalSeconds = float(split)
                        theRet["totalSeconds"] = totalSeconds

                        # theRet['secondsPerLine'] =

            # order in file will matter, there are multiple "Image Size" lines
            # we want the first
            # "Image Size"	"294 * 1000 [pixel]"
            elif line.startswith('"Image Size"'):
                if line.startswith('"Image Size(Unit Converted)"'):
                    continue
                if gotImageSize:
                    continue
                gotImageSize = True
                line = line.replace('"', "")
                line = line.replace(",", "")
                splitLine = line.split("\t")  # yes, a FREAKING tab !!!!
                splitLine = splitLine[1]
                splitLine2 = splitLine.split()
                # print('splitLine2:', splitLine2)

                theRet["numLines"] = int(splitLine2[2])

            # elif line.startswith('"Image Size(Unit Converted)"'):
            # 	print('loadLineScanHeader:', line)

    # tif shape is (lines, pixels)
    shape = (theRet["numLines"], theRet["numPixels"])
    # theRet['shape'] = self.tif.shape
    # theRet['shape'] = tifData.shape
    theRet["shape"] = shape
    theRet["secondsPerLine"] = theRet["totalSeconds"] / theRet["shape"][0]
    theRet["linesPerSecond"] = 1 / theRet["secondsPerLine"]
    #
    return theRet



"""
Display a tif kymograph and allow user to modify a rect roi

Used by main SanPy interface and kymograph plugin
"""

import sys
import numpy as np
from functools import partial

from PyQt5 import QtCore, QtWidgets, QtGui
import pyqtgraph as pg

import sanpy

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class myCustomDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Set Kymograph Scale!")

        QBtn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel

        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        hLayout = QtWidgets.QHBoxLayout()

        xScaleLabel = QtWidgets.QLabel("Seconds/Line")
        self.xScale = QtWidgets.QDoubleSpinBox()
        self.xScale.setDecimals(4)

        yScaleLabel = QtWidgets.QLabel("Microns/Pixel")
        self.yScale = QtWidgets.QDoubleSpinBox()
        self.yScale.setDecimals(4)

        hLayout.addWidget(xScaleLabel)
        hLayout.addWidget(self.xScale)

        hLayout.addWidget(yScaleLabel)
        hLayout.addWidget(self.yScale)

        self.layout = QtWidgets.QVBoxLayout()
        # message = QtWidgets.QLabel("Set Kymograph Scale")
        # self.layout.addWidget(message)

        self.layout.addLayout(hLayout)

        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def getResults(self) -> dict:
        """Get values from user input."""
        xScale = self.xScale.value()
        yScale = self.yScale.value()
        retDict = {
            "secondsPerLine": xScale,
            "umPerPixel": yScale,
        }
        return retDict


def showdialog():
    d = QtWidgets.QDialog()
    b1 = QtWidgets.QPushButton("ok", d)
    b1.move(50, 50)
    d.setWindowTitle("Set Kymograph Scale")
    d.setWindowModality(QtCore.Qt.ApplicationModal)
    d.exec_()


class kymographImage(pg.ImageItem):
    """
    Utility class to inherit and redefine some functions.
    """

    def mouseClickEvent(self, event):
        # print("Click", event.pos())
        x = event.pos().x()
        y = event.pos().y()

    def mouseDragEvent(self, event):
        return

        if event.isStart():
            print("Start drag", event.pos())
        elif event.isFinish():
            print("Stop drag", event.pos())
        else:
            print("Drag", event.pos())

    def old_hoverEvent(self, event):
        logger.info("")
        if not event.isExit():
            # the mouse is hovering over the image; make sure no other items
            # will receive left click/drag events from here.
            event.acceptDrags(pg.QtCore.Qt.LeftButton)
            event.acceptClicks(pg.QtCore.Qt.LeftButton)


class kymographWidget(QtWidgets.QWidget):
    """Display a kymograph with contrast controls."""

    signalKymographRoiChanged = QtCore.pyqtSignal(object)  # list of [l, t, r, b]
    signalSwitchToMolar = QtCore.pyqtSignal(
        object, object, object
    )  # (boolean, kd, caConc)
    signalScaleChanged = QtCore.pyqtSignal(object)  # dict with x/y scale
    signalLineSliderChanged = QtCore.pyqtSignal(object)  # int, new line selected

    def __init__(self, ba=None, parent=None):
        """
        ba: bAnalysis
        """
        super().__init__(parent)

        logger.info(f"{type(ba)}")

        bitDepth = 8
        self._minContrast = 0
        self._maxContrast = 2**bitDepth

        self.ba = ba

        self.myImageItem = None  # kymographImage
        self.myLineRoi = None
        self.myLineRoiBackground = None
        self.myColorBarItem = None

        self._buildUI()
        self._replot()

    def on_convert_to_nm_clicked(self, value):
        onOff = value == 2
        # self.detectionWidget.toggleCrosshair(onOff)
        self.signalSwitchToMolar.emit(onOff)

    def on_button_click(self, name):
        logger.info(name)
        if name == "Reset ROI":
            newRect = self.ba.fileLoader.resetKymographRect()
            # self.ba._updateTifRoi(newRect)
            self._replot()
            self.signalKymographRoiChanged.emit(newRect)  # underlying _abf has new rect

        else:
            logger.info(f"Case not taken: {name}")

    def _buildMolarLayout(self):
        """Layout to conver sum intensity (for each line scan) into molar.

        Seems a bit silly.
        """
        molarLayout = QtWidgets.QHBoxLayout()

        # todo: add checkbox to turn kn/rest calculation on off
        #            need signal
        convertToMolarCheckBox = QtWidgets.QCheckBox("Convert to Molar")
        convertToMolarCheckBox.setChecked(False)
        convertToMolarCheckBox.stateChanged.connect(self.on_convert_to_nm_clicked)
        convertToMolarCheckBox.setDisabled(True)
        molarLayout.addWidget(convertToMolarCheckBox)

        #
        # kd
        kdLabel = QtWidgets.QLabel("kd")
        molarLayout.addWidget(kdLabel)

        kdDefault = 22.1
        self.kdSpinBox = QtWidgets.QDoubleSpinBox()
        self.kdSpinBox.setMinimum(0)
        self.kdSpinBox.setMaximum(+1e6)
        self.kdSpinBox.setValue(kdDefault)
        self.kdSpinBox.setDisabled(True)
        # self.kdSpinBox.setSpecialValueText("None")
        molarLayout.addWidget(self.kdSpinBox)

        #
        # resting Ca
        restingCaLabel = QtWidgets.QLabel("Resting Ca")
        molarLayout.addWidget(restingCaLabel)

        restingCaDefault = 113.7
        self.restingCaSpinBox = QtWidgets.QDoubleSpinBox()
        self.restingCaSpinBox.setMinimum(0)
        self.restingCaSpinBox.setMaximum(+1e6)
        self.restingCaSpinBox.setValue(restingCaDefault)
        self.restingCaSpinBox.setDisabled(True)
        # self.kdSpinBox.setSpecialValueText("None")
        molarLayout.addWidget(self.restingCaSpinBox)

        return molarLayout

    def _buildControlBarLayout(self):
        # get underlying scale from sanpy.bAbfText tif header
        xScale = "None"
        yScale = "None"
        xPixels = None
        yPixels = None
        _fileName = None
        if self.ba is not None:
            xScale = self.ba._abf.tifHeader["secondsPerLine"]
            yScale = self.ba._abf.tifHeader["umPerPixel"]

            yPixels, xPixels = self.ba.tifData.shape  # numpy order is (y,x)

            _fileName = self.ba.getFileName()

        controlBarLayout = QtWidgets.QHBoxLayout()

        self._fileNameLabel = QtWidgets.QLabel(f"{_fileName}")
        controlBarLayout.addWidget(self._fileNameLabel)

        # pixels

        aLabel = QtWidgets.QLabel(f"Pixels X:")
        controlBarLayout.addWidget(aLabel)

        self.xPixelLabel = QtWidgets.QLabel(f"{xPixels}")
        controlBarLayout.addWidget(self.xPixelLabel)

        aLabel = QtWidgets.QLabel(f"Y:")
        controlBarLayout.addWidget(aLabel)

        self.yPixelLabel = QtWidgets.QLabel(f"{yPixels}")
        controlBarLayout.addWidget(self.yPixelLabel)

        # scale

        aLabel = QtWidgets.QLabel(f"Scale X:")
        controlBarLayout.addWidget(aLabel)

        self.xScaleLabel = QtWidgets.QLabel(f"{xScale}")
        controlBarLayout.addWidget(self.xScaleLabel)

        aLabel = QtWidgets.QLabel(f"Y:")
        controlBarLayout.addWidget(aLabel)

        self.yScaleLabel = QtWidgets.QLabel(f"{yScale}")
        controlBarLayout.addWidget(self.yScaleLabel)

        # reset

        buttonName = "Reset ROI"
        button = QtWidgets.QPushButton(buttonName)
        # button.setToolTip('Detect spikes using dV/dt threshold.')
        button.clicked.connect(partial(self.on_button_click, buttonName))
        controlBarLayout.addWidget(button)

        return controlBarLayout

    def _buildUI(self):
        # one row of controls and then kymograph image
        self.myVBoxLayout = QtWidgets.QVBoxLayout(self)
        self.myVBoxLayout.setAlignment(QtCore.Qt.AlignTop)

        if 0:
            molarLayout = self._buildMolarLayout()
            self.myVBoxLayout.addLayout(molarLayout)  #

        #
        controlBarLayout = self._buildControlBarLayout()
        self.myVBoxLayout.addLayout(controlBarLayout)  #

        # #
        # # display the (min, max, cursor) intensity
        # controlBarLayout0 = QtWidgets.QHBoxLayout()
        # # min
        # self.tifMinLabel = QtWidgets.QLabel('Min:')
        # controlBarLayout0.addWidget(self.tifMinLabel)
        # # max
        # self.tifMaxLabel = QtWidgets.QLabel('Max:')
        # controlBarLayout0.addWidget(self.tifMaxLabel)
        # # roi min
        # self.roiMinLabel = QtWidgets.QLabel('ROI Min:')
        # controlBarLayout0.addWidget(self.roiMinLabel)
        # # roi max
        # self.roiMaxLabel = QtWidgets.QLabel('ROI Max:')
        # controlBarLayout0.addWidget(self.roiMaxLabel)

        # background min
        # self.backgroundRoiMinLabel = QtWidgets.QLabel('Background Min:')
        # controlBarLayout0.addWidget(self.backgroundRoiMinLabel)
        # background max
        # self.backgroundRoiMaxLabel = QtWidgets.QLabel('Background Max:')
        # controlBarLayout0.addWidget(self.backgroundRoiMaxLabel)

        # # cursor
        self.tifCursorLabel = QtWidgets.QLabel("Cursor:")
        # controlBarLayout0.addWidget(self.tifCursorLabel)

        #
        # self.myVBoxLayout.addLayout(controlBarLayout0) #

        #
        # contrast sliders
        bitDepth = 8

        # min
        minContrastLayout = QtWidgets.QHBoxLayout()

        minLabel = QtWidgets.QLabel("Min")
        minContrastLayout.addWidget(minLabel)

        self.minContrastSpinBox = QtWidgets.QSpinBox()
        self.minContrastSpinBox.setMinimum(0)
        self.minContrastSpinBox.setMaximum(2**bitDepth)
        minContrastLayout.addWidget(self.minContrastSpinBox)

        minContrastSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        minContrastSlider.setMinimum(0)
        minContrastSlider.setMaximum(2**bitDepth)
        minContrastSlider.setValue(0)
        # myLambda = lambda chk, item=canvasName: self._userSelectCanvas(chk, item)
        minContrastSlider.valueChanged.connect(
            lambda val, name="min": self._onContrastSliderChanged(val, name)
        )
        minContrastLayout.addWidget(minContrastSlider)

        # image min
        self.tifMinLabel = QtWidgets.QLabel("Min:")
        minContrastLayout.addWidget(self.tifMinLabel)

        # roi min
        self.roiMinLabel = QtWidgets.QLabel("ROI Min:")
        minContrastLayout.addWidget(self.roiMinLabel)

        self.myVBoxLayout.addLayout(minContrastLayout)  #

        # max
        maxContrastLayout = QtWidgets.QHBoxLayout()

        maxLabel = QtWidgets.QLabel("Max")
        maxContrastLayout.addWidget(maxLabel)

        self.maxContrastSpinBox = QtWidgets.QSpinBox()
        self.maxContrastSpinBox.setMinimum(0)
        self.maxContrastSpinBox.setMaximum(2**bitDepth)
        maxContrastLayout.addWidget(self.maxContrastSpinBox)

        maxContrastSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        maxContrastSlider.setMinimum(0)
        maxContrastSlider.setMaximum(2**bitDepth)
        maxContrastSlider.setValue(2**bitDepth)
        maxContrastSlider.valueChanged.connect(
            lambda val, name="max": self._onContrastSliderChanged(val, name)
        )
        maxContrastLayout.addWidget(maxContrastSlider)

        # image max
        self.tifMaxLabel = QtWidgets.QLabel("Max:")
        maxContrastLayout.addWidget(self.tifMaxLabel)

        # roi max
        self.roiMaxLabel = QtWidgets.QLabel("ROI Max:")
        maxContrastLayout.addWidget(self.roiMaxLabel)

        self.myVBoxLayout.addLayout(maxContrastLayout)  #

        #
        # kymograph
        self.view = pg.GraphicsLayoutWidget()
        # self.view.show()
        # self.kymographWindow = pg.PlotWidget()

        row = 0
        colSpan = 1
        rowSpan = 1
        self.kymographPlot = self.view.addPlot(
            row=row, col=0, rowSpan=rowSpan, colSpan=colSpan
        )
        self.kymographPlot.enableAutoRange()
        # turn off x/y dragging of deriv and vm
        self.kymographPlot.setMouseEnabled(x=False, y=True)
        # hide the little 'A' button to rescale axis
        self.kymographPlot.hideButtons()
        # turn off right-click menu
        self.kymographPlot.setMenuEnabled(False)
        # hide by default
        self.kymographPlot.hide()  # show in _replot() if self.ba.isKymograph()

        # TODO: add show/hide, we do not want this in the main interface
        # vertical line to show selected line scan (adjusted/changed with slider)
        self._sliceLine = pg.InfiniteLine(pos=0, angle=90)
        # self._sliceLinesList.append(sliceLine) # keep a list of vertical slice lines so we can update all at once
        self.kymographPlot.addItem(self._sliceLine)

        self.myVBoxLayout.addWidget(self.view)

        #
        # # add scatter plot for rising/falling diameter detection
        # color = 'g'
        # symbol = 'o'
        # leftDiamScatterPlot = pg.PlotDataItem(pen=None, symbol=symbol, symbolSize=6, symbolPen=None, symbolBrush=color)
        # leftDiamScatterPlot.setData(x=[], y=[]) # start empty

        # color = 'r'
        # symbol = 'o'
        # rightDiamScatterPlot = pg.PlotDataItem(pen=None, symbol=symbol, symbolSize=6, symbolPen=None, symbolBrush=color)
        # rightDiamScatterPlot.setData(x=[], y=[]) # start empty

        # self.kymographPlot.addItem(leftDiamScatterPlot)
        # self.kymographPlot.addItem(rightDiamScatterPlot)

        #
        # 1.5) slider to step through "Line Profile"
        self._profileSlider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
        self._profileSlider.setMinimum(0)
        # TODO: cludge, fix
        _numLineScans = 0
        if self.ba is not None:
            _numLineScans = len(self.ba.fileLoader.sweepX) - 1
        self._profileSlider.setMaximum(_numLineScans)
        # self._profileSlider.setMaximum(self._kymographAnalysis.numLineScans())
        self._profileSlider.valueChanged.connect(self._on_line_slider_changed)
        self.myVBoxLayout.addWidget(self._profileSlider)

    def showLineSlider(self, visible):
        """Toggle line slider and vertical line (on image) on/off."""
        self._profileSlider.setVisible(visible)

    def _on_line_slider_changed(self, lineNumber: int):
        """Respond to user dragging the line slider.

        Args:
            lineNumber: Int line number, needs to be converted to 'seconds'
        """

        lineSeconds = lineNumber * self.ba._abf.tifHeader["secondsPerLine"]

        logger.info(f"lineNumber:{lineNumber} lineSeconds:{lineSeconds}")

        # set the vertical line
        self._sliceLine.setValue(lineSeconds)

        self.signalLineSliderChanged.emit(lineNumber)

    def _onContrastSliderChanged(self, val: int, name: str):
        """Respond to either the min or max contrast slider.

        Args:
            val: new value
            name: Name of slider, in ('min', 'max')
        """
        logger.info(f"{name} {val}")
        if name == "min":
            self._minContrast = val
            self.minContrastSpinBox.setValue(val)
        elif name == "max":
            self._maxContrast = val
            self.maxContrastSpinBox.setValue(val)
        self._replot()

    def getContrastEnhance(self, theMin=None, theMax=None):
        """Get contrast enhanced image."""
        bitDepth = 8

        if theMin is None:
            theMin = self._minContrast
        if theMax is None:
            theMax = self._maxContrast

        lut = np.arange(2**bitDepth, dtype="uint8")
        lut = self._getContrastedImage(lut, theMin, theMax)  # get a copy of the image
        theRet = np.take(lut, self.ba.tifData)
        return theRet

    def _getContrastedImage(
        self, image, display_min, display_max
    ):  # copied from Bi Rico
        # Here I set copy=True in order to ensure the original image is not
        # modified. If you don't mind modifying the original image, you can
        # set copy=False or skip this step.
        bitDepth = 8

        image = np.array(image, dtype=np.uint8, copy=True)
        image.clip(display_min, display_max, out=image)
        image -= display_min
        np.floor_divide(
            image,
            (display_max - display_min + 1) / (2**bitDepth),
            out=image,
            casting="unsafe",
        )
        # np.floor_divide(image, (display_max - display_min + 1) / 256,
        #                out=image, casting='unsafe')
        # return image.astype(np.uint8)
        return image

    def _replot(self, startSec=None, stopSec=None):
        logger.info("")

        if self.ba is None:
            return

        self.kymographPlot.clear()
        self.kymographPlot.show()

        myTif = self.getContrastEnhance()

        logger.info(
            f"  startSec:{startSec} stopSec:{stopSec} myTif.shape:{myTif.shape}"
        )  # like (519, 10000)

        # self.myImageItem = pg.ImageItem(myTif, axisOrder='row-major')
        #  TODO: set height to micro-meters
        axisOrder = "row-major"

        # todo: get from tifHeader
        umLength = self.ba.tifData.shape[0] * self.ba._abf.tifHeader["umPerPixel"]

        rect = [0, 0, self.ba.fileLoader.recordingDur, umLength]  # x, y, w, h

        if self.myImageItem is None:
            # first time build
            self.myImageItem = kymographImage(myTif, axisOrder=axisOrder, rect=rect)
            # redirect hover to self (to display intensity
            self.myImageItem.hoverEvent = self.hoverEvent
        else:
            # second time update
            # myTif = self.ba.tifData
            myTif = self.getContrastEnhance()
            self.myImageItem.setImage(myTif, axisOrder=axisOrder, rect=rect)
        self.kymographPlot.addItem(self.myImageItem)

        padding = 0
        if startSec is not None and stopSec is not None:
            self.kymographPlot.setXRange(
                startSec, stopSec, padding=padding
            )  # row major is different

        # re-add the vertical line
        self.kymographPlot.addItem(self._sliceLine)

        #
        # plot of diameter detection
        self._leftFitScatter = pg.ScatterPlotItem(
            size=3, brush=pg.mkBrush(50, 255, 50, 120)
        )
        _xFake = []  # self.ba.sweepX
        _yFake = []  # [200 for x in self.ba.sweepX]
        self._leftFitScatter.setData(_xFake, _yFake)
        self.kymographPlot.addItem(self._leftFitScatter)

        self._rightFitScatter = pg.ScatterPlotItem(
            size=3, brush=pg.mkBrush(255, 50, 50, 120)
        )
        _xFake = []  # self.ba.sweepX
        _yFake = []  # [350 for x in self.ba.sweepX]
        self._rightFitScatter.setData(_xFake, _yFake)
        self.kymographPlot.addItem(self._rightFitScatter)

        #
        # color bar with contrast !!!
        if myTif.dtype == np.dtype("uint8"):
            bitDepth = 8
        elif myTif.dtype == np.dtype("uint16"):
            bitDepth = 16
        else:
            bitDepth = 16
            logger.error(f"Did not recognize tif dtype: {myTif.dtype}")

        cm = pg.colormap.get(
            "Greens_r", source="matplotlib"
        )  # prepare a linear color map
        # values = (0, 2**bitDepth)
        # values = (0, maxTif)
        values = (0, 2**12)
        limits = (0, 2**12)
        # logger.info(f'color bar bit depth is {bitDepth} with values in {values}')
        doColorBar = False
        if doColorBar:
            if self.myColorBarItem == None:
                self.myColorBarItem = pg.ColorBarItem(
                    values=values,
                    limits=limits,
                    interactive=True,
                    label="",
                    cmap=cm,
                    orientation="horizontal",
                )
            # Have ColorBarItem control colors of img and appear in 'plot':
            self.myColorBarItem.setImageItem(
                self.myImageItem, insert_in=self.kymographPlot
            )
            self.myColorBarItem.setLevels(values=values)

        kymographRect = self.ba.fileLoader.getKymographRect()
        if kymographRect is not None:
            # TODO: I guess we always have a rect, o.w. this would be a runtime error
            xRoiPos = kymographRect[0]
            yRoiPos = kymographRect[3]
            top = kymographRect[1]
            right = kymographRect[2]
            bottom = kymographRect[3]
            widthRoi = right - xRoiPos + 1
            # heightRoi = bottom - yRoiPos + 1
            heightRoi = top - yRoiPos + 1
        """
        else:
            #  TODO: Put this logic into function in bAbfText
            pos, size = self.ba.defaultTifRoi()

            xRoiPos = 0  # startSeconds
            yRoiPos = 0  # pixels
            widthRoi = myTif.shape[1]
            heightRoi = myTif.shape[0]
            tifHeightPercent = myTif.shape[0] * 0.2
            #print('tifHeightPercent:', tifHeightPercent)
            yRoiPos += tifHeightPercent
            heightRoi -= 2 * tifHeightPercent
        """
        # TODO: get this out of replot, recreating the ROI is causing runtime error
        # update the rect roi
        pos = (xRoiPos, yRoiPos)
        size = (widthRoi, heightRoi)
        if self.myLineRoi is None:
            movable = False
            self.myLineRoi = pg.ROI(
                pos=pos, size=size, parent=self.myImageItem, movable=movable
            )
            # self.myLineRoi.addScaleHandle((0,0), (1,1), name='topleft')  # at origin
            self.myLineRoi.addScaleHandle(
                (0.5, 0), (0.5, 1), name="top center"
            )  # top center
            self.myLineRoi.addScaleHandle(
                (0.5, 1), (0.5, 0), name="bottom center"
            )  # bottom center
            # self.myLineRoi.addScaleHandle((0,0.5), (1,0.5))  # left center
            # self.myLineRoi.addScaleHandle((1,0.5), (0,0.5))  # right center
            # self.myLineRoi.addScaleHandle((1,1), (0,0), name='bottomright')  # bottom right
            self.myLineRoi.sigRegionChangeFinished.connect(self.kymographChanged)
        else:
            self.myLineRoi.setPos(pos, finish=False)
            self.myLineRoi.setSize(size, finish=False)

        #
        # background kymograph ROI
        backgroundRect = (
            self.ba.fileLoader.getKymographBackgroundRect()
        )  # keep this in the backend
        if backgroundRect is not None:
            xRoiPos = backgroundRect[0]
            yRoiPos = backgroundRect[3]
            top = backgroundRect[1]
            right = backgroundRect[2]
            bottom = backgroundRect[3]
            widthRoi = right - xRoiPos + 1
            # heightRoi = bottom - yRoiPos + 1
            heightRoi = top - yRoiPos + 1
        pos = (xRoiPos, yRoiPos)
        size = (widthRoi, heightRoi)

        if 0:
            if self.myLineRoiBackground is None:
                # TODO: get this out of replot, recreating the ROI is causing runtime error
                self.myLineRoiBackground = pg.ROI(
                    pos=pos, size=size, parent=self.myImageItem
                )
            else:
                self.myLineRoiBackground.setPos(pos, finish=False)
                self.myLineRoiBackground.setSize(size, finish=False)

        # update min/max labels
        # TODO: only set this once on switch file
        myTifOrig = self.ba.fileLoader.tifData
        minTif = np.nanmin(myTifOrig)
        maxTif = np.nanmax(myTifOrig)
        # print(type(dtype), dtype)  # <class 'numpy.dtype[uint16]'> uint16
        self.tifMinLabel.setText(f"Min:{minTif}")
        self.tifMaxLabel.setText(f"Max:{maxTif}")

        # update min/max displayed to user
        self._updateRoiMinMax(kymographRect)
        self._updateBackgroundRoiMinMax(backgroundRect)

    def updateLeftRightFit(self, yLeft, yRight, visible=True):
        """ """

        self._leftFitScatter.setVisible(visible)
        self._rightFitScatter.setVisible(visible)

        if visible:
            x = self.ba.fileLoader.sweepX
            # TODO: check that len(y) == len(x)
            self._leftFitScatter.setData(x, yLeft)
            self._rightFitScatter.setData(x, yRight)

    def _updateBackgroundRoiMinMax(self, backgroundRect=None):
        """
        update background roi

        TODO: Add self.ba.fileLoader.getBackGroundStats()

        """
        logger.warning(f"Need to add interface for user to adjust background roi")
        return

        if backgroundRect is None:
            backgroundRect = self.ba.fileLoader.getKymographBackgroundRect()

        left = backgroundRect[0]
        top = backgroundRect[1]
        right = backgroundRect[2]
        bottom = backgroundRect[3]

        myTif = self.ba.tifData
        tifClip = myTif[bottom:top, left:right]

        roiMin = np.nanmin(tifClip)
        roiMax = np.nanmax(tifClip)

        # self.backgroundRoiMinLabel.setText(f'Background Min:{roiMin}')
        # self.backgroundRoiMaxLabel.setText(f'Background Max:{roiMax}')

    def _updateRoiMinMax(self, theRect):
        left = theRect[0]
        top = theRect[1]
        right = theRect[2]
        bottom = theRect[3]

        logger.info(f"left:{left} top:{top} right:{right} bottom:{bottom}")

        myTif = self.ba.fileLoader.tifData
        tifClip = myTif[bottom:top, left:right]

        roiMin = np.nanmin(tifClip)
        roiMax = np.nanmax(tifClip)

        self.roiMinLabel.setText(f"ROI Min:{roiMin}")
        self.roiMaxLabel.setText(f"ROI Max:{roiMax}")

    def kymographChanged(self, event):
        """
        User finished gragging the ROI

        Args:
            event (pyqtgraph.graphicsItems.ROI.ROI)
        """
        logger.info("")
        # pos = event.pos()
        # size = event.size()

        _kymographRect = self.ba.fileLoader.getKymographRect()  # (l, t, r, b)

        left = _kymographRect[0]
        top = _kymographRect[1]
        right = _kymographRect[2]
        bottom = _kymographRect[3]

        handles = event.getSceneHandlePositions()
        for _idx, handle in enumerate(handles):
            logger.info(f"{_idx} handle: {handle}")
            if handle[0] is not None:
                # imagePos = self.myImageItem.mapFromScene(handle[1])
                imagePos = self.myImageItem.mapFromScene(handle[1])
                x = imagePos.x()
                y = imagePos.y()
                # units are in image pixels !!!
                # if handle[0] == 'topleft':
                #     left = x
                #     bottom = y
                # elif handle[0] == 'bottomright':
                #     right = x
                #     top = y
                if handle[0] == "top center":
                    top = y
                elif handle[0] == "bottom center":
                    bottom = y

        #
        left = int(left)
        top = int(top)
        right = int(right)
        bottom = int(bottom)

        if left < 0:
            left = 0
        if bottom < 0:
            bottom = 0

        # force left and right
        _kymographRect = self.ba.fileLoader.getKymographRect()  # (l, t, r, b)
        left = 0
        right = _kymographRect[2]

        logger.info(f"  left:{left} top:{top} right:{right} bottom:{bottom}")

        #  cludge
        if bottom > top:
            logger.warning(f"fixing bad top/bottom")
            tmp = top
            top = bottom
            bottom = tmp

        theRect = [left, top, right, bottom]

        self._updateRoiMinMax(theRect)

        # TODO: detection widget needs a slot to (i) analyze and then replot
        # self.ba._updateTifRoi(theRect)
        # self._replot(startSec=None, stopSec=None, userUpdate=True)
        # self.signalDetect.emit(self.ba)  # underlying _abf has new rect
        self.ba.fileLoader._updateTifRoi(theRect)
        self.signalKymographRoiChanged.emit(theRect)  # underlying _abf has new rect

    def slot_switchFile(self, ba=None, startSec=None, stopSec=None):
        if ba is not None and ba.fileLoader.isKymograph():
            self.ba = ba
            # self._updateRoiMinMax(theRect)
            self._replot(startSec=startSec, stopSec=stopSec)

            # update line scan slider
            _numLineScans = len(self.ba.fileLoader.sweepX) - 1
            self._profileSlider.setMaximum(_numLineScans)

    def hoverEvent(self, event):
        if event.isExit():
            return

        xPos = event.pos().x()
        yPos = event.pos().y()

        xPos = int(xPos)
        yPos = int(yPos)

        myTif = self.ba.fileLoader.tifData
        intensity = myTif[yPos, xPos]  # flipped

        # logger.info(f'x:{xPos} y:{yPos} intensity:{intensity}')

        self.tifCursorLabel.setText(f"Cursor:{intensity}")
        self.tifCursorLabel.update()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    if 1:
        path = "/media/cudmore/data/rabbit-ca-transient/jan-12-2022/Control/220110n_0003.tif.frames/220110n_0003.tif"
        path = "/Users/cudmore/data/rosie/test-data/filter median 1 C2-0-255 Cell 2 CTRL  2_5_21 female wt old.tif"

        ba = sanpy.bAnalysis(path)
        print(ba)

        kw = kymographWidget(ba)
        kw.show()

    # test dialog
    if 0:
        mcd = myCustomDialog()
        if mcd.exec():
            scaleDict = mcd.getResults()
            print(scaleDict)
        else:
            print('user did not hit "ok"')

    sys.exit(app.exec_())



import numpy as np
import scipy.signal

from PyQt5 import QtCore, QtWidgets, QtGui

import matplotlib.pyplot as plt

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.bAnalysisUtil import statList


class exportTrace(sanpyPlugin):
    """ """

    myHumanName = "Export Trace"

    def __init__(self, **kwargs):
        super(exportTrace, self).__init__(**kwargs)

        # this plugin will not respond to any interface changes
        # self.turnOffAllSignalSlot()

        self.mainWidget = None

        self.plot()

    def replot(self):
        logger.info(f"sweepNumber: {self.sweepNumber}")

        if self.mainWidget is None:
            self.plot()

        x = self.getSweep("x")  # get for current sweep
        y = self.getSweep("y")
        self.mainWidget.switchFile(self.ba, x, y)

    def plot(self):
        if self.ba is None:
            return

        self.myType = "vmFiltered"

        if self.myType == "vmFiltered":
            xyUnits = ("Time (sec)", "Vm (mV)")  # todo: pass xMin,xMax to constructor
            x = self.getSweep("x")  # corresponds to sweepX
            y = self.getSweep("y")  # corresponds to sweepY
        elif self.myType == "dvdtFiltered":
            xyUnits = (
                "Time (sec)",
                "dV/dt (mV/ms)",
            )  # todo: pass xMin,xMax to constructor
        elif self.myType == "meanclip":
            xyUnits = ("Time (ms)", "Vm (mV)")  # todo: pass xMin,xMax to constructor
        else:
            logger.error(f'Unknown myType: "{self.myType}"')
            xyUnits = ("error time", "error y")

        path = self.ba.fileLoader.filepath

        xMin, xMax = self.getStartStop()
        """
        xMin = None
        xMax = None
        if self.myType in ['clip', 'meanclip']:
            xMin, xMax = self.detectionWidget.clipPlot.getAxis('bottom').range
        else:
            xMin, xMax = self.detectionWidget.getXRange()
        """

        if self.myType in ["vm", "dvdt"]:
            xMargin = 2  # seconds
        else:
            xMargin = 2

        # a large x margin was causing matplotlib erors when a swep was short, like 0.2 sec
        xMargin = 0

        tmpLayout = QtWidgets.QVBoxLayout()

        self.mainWidget = sanpy.interface.bExportWidget(
            x,
            y,
            xyUnits=xyUnits,
            path=path,
            xMin=xMin,
            xMax=xMax,
            xMargin=xMargin,
            type=self.myType,
            darkTheme=True,
        )
        # darkTheme=self.detectionWidget.useDarkStyle)

        tmpLayout.addWidget(self.mainWidget)

        # rewire existing widget into plugin architecture
        # self.mainWidget.closeEvent = self.onClose
        # self._mySetWindowTitle()

        # self.setCentralWidget(self.mainWidget)
        # self.setLayout(tmpLayout)
        self.getVBoxLayout().addLayout(tmpLayout)


if __name__ == "__main__":
    path = "/Users/cudmore/Sites/SanPy/data/19114001.abf"
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()
    print(ba.numSpikes)

    import sys

    app = QtWidgets.QApplication([])
    et = exportTrace(ba=ba)
    et.show()

    path2 = "/Users/cudmore/Sites/SanPy/data/19114000.abf"
    ba2 = sanpy.bAnalysis(path2)
    et.mainWidget.switchFile(ba2)

    sys.exit(app.exec_())
